@@ -35,10 +35,8 @@ public:
 
   // XXX: instead in systemtap_session?
   void derive_probes (match_node * root, probe *p, std::vector<derived_probe*>& dps);
-
-protected:
   
-  vardecl* find_var (const std::string& name, unsigned arity);
+  vardecl* find_var (const std::string& name, int arity);
   functiondecl* find_function (const std::string& name, unsigned arity);
 
   void visit_block (block *s);
@@ -46,6 +44,7 @@ protected:
   void visit_foreach_loop (foreach_loop* e);
   void visit_arrayindex (arrayindex* e);
   void visit_functioncall (functioncall* e);
+  void visit_delete_statement (delete_statement* s);
 };
 
 
@@ -91,6 +90,7 @@ struct typeresolution_info: public visitor
   void visit_ternary_expression (ternary_expression* e);
   void visit_assignment (assignment* e);
   void visit_symbol (symbol* e);
+  void visit_target_symbol (target_symbol* e);
   void visit_arrayindex (arrayindex* e);
   void visit_functioncall (functioncall* e);
 };
@@ -137,8 +137,7 @@ struct derived_probe: public probe
 
 // ------------------------------------------------------------------------
 
-struct
-derived_probe_builder
+struct derived_probe_builder
 {
   virtual void build(systemtap_session & sess,
 		     probe* base, 
@@ -147,6 +146,11 @@ derived_probe_builder
 		     std::vector<probe*> & results_to_expand_further,
 		     std::vector<derived_probe*> & finished_results) = 0;
   virtual ~derived_probe_builder() {}
+
+  static bool get_param (std::map<std::string, literal*> const & parameters,
+                         const std::string& key, std::string& value);
+  static bool get_param (std::map<std::string, literal*> const & parameters,
+                         const std::string& key, int64_t& value);
 };
 
 
@@ -227,6 +231,9 @@ struct systemtap_session
   std::vector<derived_probe*> probes;
   std::vector<embeddedcode*> embeds;
 
+  // module-referencing file handles
+  std::map<std::string,int> module_fds;
+
   // unparser data
   translator_output* op;
   unparser* up;
