@@ -30,14 +30,21 @@ using namespace std;
 
 
 void
-usage (systemtap_session& s)
+version ()
 {
   clog
     << "SystemTap translator/driver "
     << "(version " << VERSION << " built " << DATE << ")" << endl
     << "Copyright (C) 2005 Red Hat, Inc." << endl
     << "This is free software; see the source for copying conditions."
-    << endl
+    << endl;
+}
+
+void
+usage (systemtap_session& s)
+{
+  version ();
+  clog
     << endl
     << "Usage: stap [options] FILE         Run script in file."
     << endl
@@ -46,10 +53,11 @@ usage (systemtap_session& s)
     << "   or: stap [options] -e SCRIPT    Run given script."
     << endl
     << endl
-    << "Arguments:" << endl
+    << "Options:" << endl
     << "   --         no more options after this" << endl
-    << "   -v         verbose" << (s.verbose ? " [set]" : "")
-    << endl
+    << "   -v         verbose" << (s.verbose ? " [set]" : "") << endl
+    << "   -h         show help" << endl
+    << "   -V         show version" << endl
     << "   -k         keep temporary directory" << endl
     // << "   -t         test mode" << (s.test_mode ? " [set]" : "") << endl
     << "   -g         guru mode" << (s.guru_mode ? " [set]" : "") << endl
@@ -123,11 +131,15 @@ main (int argc, char * const argv [])
 
   while (true)
     {
-      int grc = getopt (argc, argv, "vp:I:e:o:tR:r:m:kg");
+      int grc = getopt (argc, argv, "hVvp:I:e:o:tR:r:m:kg");
       if (grc < 0)
         break;
       switch (grc)
         {
+        case 'V':
+          version ();
+          exit (0);
+
         case 'v':
 	  s.verbose = true;
 	  break;
@@ -180,7 +192,6 @@ main (int argc, char * const argv [])
           s.guru_mode = true;
           break;
 
-        case '?':
         case 'h':
         default:
           usage (s);
@@ -295,11 +306,12 @@ main (int argc, char * const argv [])
       cout << "# parse tree dump" << endl;
       s.user_file->print (cout);
       cout << endl;
-      for (unsigned i=0; i<s.library_files.size(); i++)
-        {
-          s.library_files[i]->print (cout);
-          cout << endl;
-        }
+      if (s.verbose)
+        for (unsigned i=0; i<s.library_files.size(); i++)
+          {
+            s.library_files[i]->print (cout);
+            cout << endl;
+          }
     }
 
   // syntax errors, if any, are already printed
@@ -398,6 +410,7 @@ main (int argc, char * const argv [])
     cerr << "Pass 4: compilation failed.  "
          << "Try again with '-v' (verbose) option." << endl;
 
+  // XXX: what to do if rc==0 && last_pass == 4?  dump .ko file to stdout? 
   if (rc || s.last_pass == 4) goto cleanup;
 
   // PASS 5: RUN
