@@ -55,17 +55,25 @@ parser::parse (const std::string& n, bool pr)
   return p.parse ();
 }
 
+static string
+tt2str(token_type tt)
+{
+  switch (tt)
+    {
+    case tok_junk: return "junk";
+    case tok_identifier: return "identifier";
+    case tok_operator: return "operator";
+    case tok_string: return "string";
+    case tok_number: return "number";
+    case tok_embedded: return "embedded-code";
+    }
+  return "unknown token";
+}
 
 ostream&
 operator << (ostream& o, const token& t)
 {
-  o << (t.type == tok_junk ? "junk" :
-        t.type == tok_identifier ? "identifier" :
-        t.type == tok_operator ? "operator" :
-        t.type == tok_string ? "string" :
-        t.type == tok_number ? "number" :
-        t.type == tok_embedded ? "embedded-code" :
-        "unknown token");
+  o << tt2str(t.type);
 
   if (t.type != tok_embedded) // XXX: other types?
     {
@@ -141,6 +149,70 @@ parser::peek ()
 }
 
 
+static inline bool
+tok_is(token const * t, token_type tt, string const & expected)
+{
+  return t && t->type == tt && t->content == expected;
+}
+
+
+const token* 
+parser::expect_known (token_type tt, string const & expected)
+{
+  const token *t = next();
+  if (! t && t->type == tt && t->content == expected)
+    throw parse_error ("expected '" + expected + "'");
+  return t;
+}
+
+
+const token* 
+parser::expect_unknown (token_type tt, string & target)
+{
+  const token *t = next();
+  if (!(t && t->type == tt))
+    throw parse_error ("expected " + tt2str(tt));
+  target = t->content;
+  return t;
+}
+
+
+const token* 
+parser::expect_op (std::string const & expected)
+{
+  return expect_known (tok_operator, expected);
+}
+
+
+const token* 
+parser::expect_kw (std::string const & expected)
+{
+  return expect_known (tok_identifier, expected);
+}
+
+
+const token* 
+parser::expect_ident (std::string & target)
+{
+  return expect_unknown (tok_identifier, target);
+}
+
+
+bool 
+parser::peek_op (std::string const & op)
+{
+  return tok_is (peek(), tok_operator, op);
+}
+
+
+bool 
+parser::peek_kw (std::string const & kw)
+{
+  return tok_is (peek(), tok_identifier, kw);
+}
+
+
+
 lexer::lexer (istream& i, const string& in):
   input (i), input_name (in), cursor_line (1), cursor_column (1)
 { }
@@ -219,7 +291,7 @@ lexer::scan ()
       return n;
     }
 
-  else if (isdigit (c))
+  else if (isdigit (c)) // positive literal
     {
       n->type = tok_number;
       n->content = (char) c;
@@ -277,6 +349,11 @@ lexer::scan ()
       string s2 = (c2 > 0 ? s1 + (char) c2 : s1);
       string s3 = (c3 > 0 ? s2 + (char) c3 : s2);
 
+      // NB: if we were to recognize negative numeric literals here,
+      // we'd introduce another grammar ambiguity:
+      // 1-1 would be parsed as tok_number(1) and tok_number(-1)
+      // instead of tok_number(1) tok_operator('-') tok_number(1)
+
       if (s1 == "#") // shell comment
         {
           unsigned this_line = cursor_line;
@@ -775,15 +852,17 @@ parser::parse_literal ()
       const char* startp = t->content.c_str ();
       char* endp = (char*) startp;
 
-      // NB: we allow controlled overflow from LONG_MIN .. ULONG_MAX
+      // NB: we allow controlled overflow from LLONG_MIN .. ULLONG_MAX
+      // Actually, this allows all the way from -ULLONG_MAX to ULLONG_MAX,
+      // since the lexer only gives us positive digit strings.
       errno = 0;
-      long long value = strtoll (startp, & endp, 0);
+      long long value = (long long) strtoull (startp, & endp, 0);
       if (errno == ERANGE || errno == EINVAL || *endp != '\0'
-          || value > 4294967295LL || value < (-2147483647LL-1))
+          || (unsigned long long) value > 18446744073709551615ULL
+          || value < -9223372036854775807LL-1)
         throw parse_error ("number invalid or out of range"); 
 
-      long value2 = (long) value;
-      l = new literal_number (value2);
+      l = new literal_number (value);
     }
   else
     throw parse_error ("expected literal string or number");
@@ -1465,7 +1544,7 @@ parser::parse_unary ()
       e->op = t->content;
       e->tok = t;
       next ();
-      e->operand = parse_expression ();
+      e->operand = parse_crement ();
       return e;
     }
   else
@@ -1534,41 +1613,48 @@ parser::parse_value ()
 expression*
 parser::parse_symbol () 
 {
-  const token* t = next ();
-  if (t->type != tok_identifier)
-    throw parse_error ("expected identifier");
+  string name;
+  const token* t = expect_ident (name);
   const token* t2 = t;
-  string name = t->content;
   
-  t = peek ();
-  if (t && t->type == tok_operator && t->content == "->")
+  if (name.size() > 0 && name[0] == '$')
     {
-      // shorthand for process- or thread-specific array element
-      // map "thread->VAR" to "VAR[$tid]",
-      // and "process->VAR" to "VAR[$pid]"
-      symbol* sym = new symbol;
-      if (name == "thread")
-        sym->name = "$tid";
-      else if (name == "process") 
-        sym->name = "$pid";
-      else 
-        throw parse_error ("expected 'thread->' or 'process->'");
-      struct token* t2prime = new token (*t2);
-      t2prime->content = sym->name;
-      sym->tok = t2prime;
-
-      next (); // swallow "->"
-      t = next ();
-      if (! (t->type == tok_identifier))
-        throw parse_error ("expected identifier");
-
-      struct arrayindex* ai = new arrayindex;
-      ai->tok = t;
-      ai->base = t->content;
-      ai->indexes.push_back (sym);
-      return ai;
+      // target_symbol time
+      target_symbol *tsym = new target_symbol;
+      tsym->tok = t;
+      tsym->base_name = name;
+      while (true)
+	{
+	  string c;
+	  if (peek_op ("."))
+	    { 
+	      next();
+	      expect_ident (c);
+	      tsym->components.push_back
+		(make_pair (target_symbol::comp_struct_member, c));
+	    }
+	  else if (peek_op ("->"))
+	    { 
+	      next(); 
+	      expect_ident (c);
+	      tsym->components.push_back
+		(make_pair (target_symbol::comp_struct_pointer_member, c));
+	    }
+	  else if (peek_op ("["))
+	    { 
+	      next();
+	      expect_unknown (tok_number, c);
+	      expect_op ("]");
+	      tsym->components.push_back
+		(make_pair (target_symbol::comp_literal_array_index, c));
+	    }	    
+	  else
+	    break;
+	}
+      return tsym;
     }
-  else if (t && t->type == tok_operator && t->content == "[") // array
+  
+  if (peek_op ("[")) // array
     {
       next ();
       struct arrayindex* ai = new arrayindex;
@@ -1577,25 +1663,29 @@ parser::parse_symbol ()
       while (1)
         {
           ai->indexes.push_back (parse_expression ());
-          t = next ();
-          if (t->type == tok_operator && t->content == "]")
-            break;
-          if (t->type == tok_operator && t->content == ",")
-            continue;
+          if (peek_op ("]"))
+            { 
+	      next(); 
+	      break; 
+	    }
+          else if (peek_op (","))
+	    {
+	      next();
+	      continue;
+	    }
           else
             throw parse_error ("expected ',' or ']'");
         }
       return ai;
     }
-  else if (t && t->type == tok_operator && t->content == "(") // function call
+  else if (peek_op ("(")) // function call
     {
       next ();
       struct functioncall* f = new functioncall;
       f->tok = t2;
       f->function = name;
       // Allow empty actual parameter list
-      const token* t3 = peek ();
-      if (t3 && t3->type == tok_operator && t3->content == ")")
+      if (peek_op (")"))
 	{
 	  next ();
 	  return f;
@@ -1603,11 +1693,16 @@ parser::parse_symbol ()
       while (1)
 	{
 	  f->args.push_back (parse_expression ());
-	  t = next ();
-	  if (t->type == tok_operator && t->content == ")")
-	    break;
-	  if (t->type == tok_operator && t->content == ",")
-	    continue;
+	  if (peek_op (")"))
+	    {
+	      next();
+	      break;
+	    }
+	  else if (peek_op (","))
+	    {
+	      next();
+	      continue;
+	    }
 	  else
 	    throw parse_error ("expected ',' or ')'");
 	}
