@@ -36,25 +36,29 @@
  * @param args A variable number of args in a format like printf.
  * @ingroup io
  */
+static void _stp_dbug (char *func, int line, const char *fmt, ...);
 #define dbug(args...) _stp_dbug(__FUNCTION__, __LINE__, args)
+#define kbug(args...) {printk("%s:%d ",__FUNCTION__, __LINE__); printk(args); }
 #else
 #define dbug(args...) ;
+#define kbug(args...) ;
 #endif /* DEBUG */
 
 /* atomic globals */
 static atomic_t _stp_transport_failures = ATOMIC_INIT (0);
 
-#ifndef STP_NETLINK_ONLY
+#ifdef STP_RELAYFS
 static struct
 {
 	atomic_t ____cacheline_aligned_in_smp seq;
 } _stp_seq = { ATOMIC_INIT (0) };
 
 #define _stp_seq_inc() (atomic_inc_return(&_stp_seq.seq))
-#endif
+#endif /* RELAYFS */
 
 #include "print.c"
 #include "string.c"
+#include "arith.c"
 
 /************* Module Stuff ********************/
 int probe_start(void);
