@@ -106,7 +106,8 @@ vardecl::vardecl ():
 void
 vardecl::set_arity (int a)
 {
-  assert (a >= 0);
+  if (a < 0)
+    return;
 
   if (arity != a && arity >= 0)
     throw semantic_error ("inconsistent arity", tok);
@@ -123,7 +124,7 @@ vardecl::set_arity (int a)
 bool 
 vardecl::compatible_arity (int a)
 {
-  if (arity == -1)
+  if (arity == -1 || a == -1)
     return true;
   return arity == a;
 }
@@ -135,7 +136,7 @@ functiondecl::functiondecl ():
 }
 
 
-literal_number::literal_number (long v)
+literal_number::literal_number (int64_t v)
 {
   value = v;
   type = pe_long;
@@ -219,8 +220,8 @@ void post_crement::print (ostream& o)
 
 void ternary_expression::print (ostream& o)
 {
-  o << "(" << *cond << ") ? ("
-    << *truevalue << ") : ("
+  o << "(" << *cond << ")?("
+    << *truevalue << "):("
     << *falsevalue << ")"; 
 }
 
@@ -231,6 +232,26 @@ void symbol::print (ostream& o)
 }
 
 
+void target_symbol::print (std::ostream& o)
+{
+  o << base_name;
+  for (unsigned i = 0; i < components.size(); ++i)
+    {
+      switch (components[i].first)
+	{
+	case comp_literal_array_index:
+	  o << '[' << components[i].second << ']';
+	  break;
+	case comp_struct_pointer_member:
+	  o << "->" << components[i].second;
+	  break;
+	case comp_struct_member:
+	  o << "." << components[i].second;
+	}
+    }
+}
+
+
 void vardecl::print (ostream& o)
 {
   o << name;
@@ -539,7 +560,9 @@ return_statement::visit (visitor* u)
 void
 delete_statement::visit (visitor* u)
 {
+  u->push_active_lvalue (this->value);
   u->visit_delete_statement (this);
+  u->pop_active_lvalue ();
 }
 
 void
@@ -593,13 +616,17 @@ unary_expression::visit (visitor* u)
 void
 pre_crement::visit (visitor* u)
 {
+  u->push_active_lvalue (this->operand);
   u->visit_pre_crement (this);
+  u->pop_active_lvalue ();
 }
 
 void
 post_crement::visit (visitor* u)
 {
+  u->push_active_lvalue (this->operand);
   u->visit_post_crement (this);
+  u->pop_active_lvalue ();
 }
 
 void
@@ -641,7 +668,9 @@ ternary_expression::visit (visitor* u)
 void
 assignment::visit (visitor* u)
 {
+  u->push_active_lvalue (this->left);
   u->visit_assignment (this);
+  u->pop_active_lvalue ();
 }
 
 void
@@ -650,6 +679,12 @@ symbol::visit (visitor* u)
   u->visit_symbol (this);
 }
 
+void 
+target_symbol::visit (visitor* u)
+{
+  u->visit_target_symbol(this);
+}
+
 void
 arrayindex::visit (visitor* u)
 {
@@ -662,6 +697,33 @@ functioncall::visit (visitor* u)
   u->visit_functioncall (this);
 }
 
+// ------------------------------------------------------------------------
+
+bool 
+visitor::is_active_lvalue(expression *e)
+{
+  for (unsigned i = 0; i < active_lvalues.size(); ++i)
+    {
+      if (active_lvalues[i] == e)
+	return true;
+    }
+  return false;
+}
+
+void 
+visitor::push_active_lvalue(expression *e)
+{
+  active_lvalues.push_back(e);
+}
+
+void 
+visitor::pop_active_lvalue()
+{
+  assert(!active_lvalues.empty());
+  active_lvalues.pop_back();
+}
+
+
 
 // ------------------------------------------------------------------------
 
@@ -831,6 +893,11 @@ traversing_visitor::visit_symbol (symbol* e)
 {
 }
 
+void
+traversing_visitor::visit_target_symbol (target_symbol* e)
+{
+}
+
 void
 traversing_visitor::visit_arrayindex (arrayindex* e)
 {
@@ -1016,6 +1083,12 @@ throwing_visitor::visit_symbol (symbol* e)
   throwone (e->tok);
 }
 
+void
+throwing_visitor::visit_target_symbol (target_symbol* e)
+{
+  throwone (e->tok);
+}
+
 void
 throwing_visitor::visit_arrayindex (arrayindex* e)
 {
@@ -1297,6 +1370,16 @@ deep_copy_visitor::visit_symbol (symbol* e)
   provide <symbol*> (this, n);
 }
 
+void
+deep_copy_visitor::visit_target_symbol (target_symbol* e)
+{
+  target_symbol* n = new target_symbol;
+  n->tok = e->tok;
+  n->base_name = e->base_name;
+  n->components = e->components;
+  provide <target_symbol*> (this, n);
+}
+
 void
 deep_copy_visitor::visit_arrayindex (arrayindex* e)
 {
