@@ -14,7 +14,9 @@
 #include <vector>
 #include <iostream>
 #include <stdexcept>
-
+extern "C" {
+#include <stdint.h>
+}
 
 struct token; // parse.h
 struct semantic_error: public std::runtime_error
@@ -38,8 +40,8 @@ struct semantic_error: public std::runtime_error
 enum exp_type
   {
     pe_unknown,
-    pe_long,
-    pe_string, 
+    pe_long,   // int64_t
+    pe_string, // std::string
     pe_stats 
   };
 
@@ -77,9 +79,8 @@ struct literal_string: public literal
 
 struct literal_number: public literal
 {
-  // XXX: s/long/long long/ throughout
-  long value;
-  literal_number (long v);
+  int64_t value;
+  literal_number (int64_t v);
   void print (std::ostream& o);
   void visit (visitor* u);
 };
@@ -177,6 +178,21 @@ struct symbol: public expression
 };
 
 
+struct target_symbol : public expression
+{
+  enum component_type
+    {
+      comp_struct_member,
+      comp_struct_pointer_member,
+      comp_literal_array_index
+    };
+  std::string base_name;
+  std::vector<std::pair<component_type, std::string> > components;
+  void print (std::ostream& o);
+  void visit (visitor* u);  
+};
+
+
 struct arrayindex: public expression
 {
   std::string base;
@@ -426,6 +442,12 @@ struct probe_alias: public probe
 // statement/expression tree.
 struct visitor
 {
+  // Machinery for differentiating lvalue visits from non-lvalue.
+  std::vector<expression *> active_lvalues;
+  bool is_active_lvalue(expression *e);
+  void push_active_lvalue(expression *e);
+  void pop_active_lvalue();
+
   virtual ~visitor () {}
   virtual void visit_block (block *s) = 0;
   virtual void visit_embeddedcode (embeddedcode *s) = 0;
@@ -453,6 +475,7 @@ struct visitor
   virtual void visit_ternary_expression (ternary_expression* e) = 0;
   virtual void visit_assignment (assignment* e) = 0;
   virtual void visit_symbol (symbol* e) = 0;
+  virtual void visit_target_symbol (target_symbol* e) = 0;
   virtual void visit_arrayindex (arrayindex* e) = 0;
   virtual void visit_functioncall (functioncall* e) = 0;
 };
@@ -489,6 +512,7 @@ struct traversing_visitor: public visitor
   void visit_ternary_expression (ternary_expression* e);
   void visit_assignment (assignment* e);
   void visit_symbol (symbol* e);
+  void visit_target_symbol (target_symbol* e);
   void visit_arrayindex (arrayindex* e);
   void visit_functioncall (functioncall* e);
 };
@@ -530,6 +554,7 @@ struct throwing_visitor: public visitor
   void visit_ternary_expression (ternary_expression* e);
   void visit_assignment (assignment* e);
   void visit_symbol (symbol* e);
+  void visit_target_symbol (target_symbol* e);
   void visit_arrayindex (arrayindex* e);
   void visit_functioncall (functioncall* e);
 };
@@ -573,6 +598,7 @@ struct deep_copy_visitor: public visitor
   virtual void visit_ternary_expression (ternary_expression* e);
   virtual void visit_assignment (assignment* e);
   virtual void visit_symbol (symbol* e);
+  virtual void visit_target_symbol (target_symbol* e);
   virtual void visit_arrayindex (arrayindex* e);
   virtual void visit_functioncall (functioncall* e);
 };
