@@ -95,6 +95,8 @@ struct c_unparser: public unparser, public visitor
 			       set< exp_type > & value_types,
 			       set< vector<exp_type> > & index_types);
 
+  void visit_statement (statement* s, unsigned actions);
+
   void visit_block (block* s);
   void visit_embeddedcode (embeddedcode* s);
   void visit_null_statement (null_statement* s);
@@ -121,6 +123,7 @@ struct c_unparser: public unparser, public visitor
   void visit_ternary_expression (ternary_expression* e);
   void visit_assignment (assignment* e);
   void visit_symbol (symbol* e);
+  void visit_target_symbol (target_symbol* e);
   void visit_arrayindex (arrayindex* e);
   void visit_functioncall (functioncall* e);
 };
@@ -139,7 +142,7 @@ struct c_tmpcounter:
     parent->tmpvar_counter = 0;
   }
 
-  // void visit_for_loop (for_loop* s);
+  void visit_for_loop (for_loop* s);
   void visit_foreach_loop (foreach_loop* s);
   // void visit_return_statement (return_statement* s);
   // void visit_delete_statement (delete_statement* s);
@@ -149,7 +152,7 @@ struct c_tmpcounter:
   void visit_post_crement (post_crement* e);
   // void visit_logical_or_expr (logical_or_expr* e);
   // void visit_logical_and_expr (logical_and_expr* e);
-  // void visit_array_in (array_in* e);
+  void visit_array_in (array_in* e);
   // void visit_comparison (comparison* e);
   void visit_concatenation (concatenation* e);
   // void visit_ternary_expression (ternary_expression* e);
@@ -235,17 +238,6 @@ public:
       return "global_" + name;
   }
 
-  void check_dbz(c_unparser &c) const
-  {
-    if (ty == pe_long)
-      {
-	c.o->newline() << "if (" << qname() << " == 0) {";
-	c.o->newline(1) << "c->errorcount++;";
-	c.o->newline() << qname() << " = 1;";
-	c.o->newline(-1) << "}";
-      }
-  }
-
   string init() const
   {
     switch (type())
@@ -280,8 +272,7 @@ struct stmt_expr
   }
   ~stmt_expr()
   {
-    c.o->line() << "})";
-    c.o->indent(-1);
+    c.o->newline(-1) << "})";
   }
 };
 
@@ -289,20 +280,32 @@ struct varlock
 {
   c_unparser & c;
   var const & v;
+  bool w;
 
-  varlock(c_unparser & c, var const & v) : c(c), v(v)
+  varlock(c_unparser & c, var const & v, bool w): c(c), v(v), w(w)
   {
-    if (!v.is_local())
-      c.o->newline() << "/* XXX lock " << v << " */";
+    if (v.is_local()) return;
+    c.o->newline() << (w ? "write_lock" : "read_lock")
+                   << " (& " << v << "_lock);";
   }
 
   ~varlock()
   {
-    if (!v.is_local())
-      c.o->newline() << "/* XXX unlock " << v << " */";
+    if (v.is_local()) return;
+    c.o->newline() << (w ? "write_unlock" : "read_unlock")
+                   << " (& " << v << "_lock);";
   }
 };
 
+struct varlock_r: public varlock {
+  varlock_r (c_unparser& c, var const& v): varlock (c, v, false) {}
+};
+
+struct varlock_w: public varlock {
+  varlock_w (c_unparser& c, var const& v): varlock (c, v, true) {}
+};
+
+
 struct tmpvar
   : public var
 {
@@ -327,6 +330,16 @@ struct mapvar
   static string key_typename(exp_type e);
   static string value_typename(exp_type e);
 
+  string del () const
+  {
+    return "_stp_map_key_del (" + qname() + ")";
+  }
+
+  string exists () const
+  {
+    return "_stp_map_entry_exists (" + qname() + ")";
+  }
+
   string seek (vector<tmpvar> const & indices) const
   {
     string result = "_stp_map_key" + mangled_indices() + " (";
@@ -344,13 +357,25 @@ struct mapvar
 
   string get () const
   {
-    return "_stp_map_get_" + shortname(type()) + " (" + qname() + ")";
+    // see also itervar::get_key
+    if (type() == pe_string)
+        // impedance matching: NULL -> empty strings
+      return "({ char *v = "
+        "_stp_map_get_" + shortname(type()) + " (" + qname() + "); "
+        "if (!v) v = \"\"; v; })";
+    else // long?
+      return "_stp_map_get_" + shortname(type()) + " (" + qname() + ")";
   }
 
   string set (tmpvar const & tmp) const
   {
-    return ("_stp_map_set_" + shortname(type()) + " (" + qname()
-	    + ", " + tmp.qname() + ")");
+    // impedance matching: empty strings -> NULL
+    if (type() == pe_string)
+      return ("_stp_map_set_" + shortname(type()) + " (" + qname()
+              + ", (" + tmp.qname() + "[0] ? " + tmp.qname() + " : NULL))");
+    else
+      return ("_stp_map_set_" + shortname(type()) + " (" + qname()
+              + ", " + tmp.qname() + ")");
   }
 
   string mangled_indices() const
@@ -420,12 +445,18 @@ public:
   
   string get_key (exp_type ty, unsigned i) const
   {
+    // bug translator/1175: runtime uses base index 1 for the first dimension
+    // see also mapval::get
     switch (ty)
       {
       case pe_long:
-	return "_stp_key_get_int64 ("+ qname() + ", " + stringify(i) + ")";
+	return "_stp_key_get_int64 ("+ qname() + ", " + stringify(i+1) + ")";
       case pe_string:
-	return "_stp_key_get_str ("+ qname() + ", " + stringify(i) + ")";
+        // impedance matching: NULL -> empty strings
+	return "({ char *v = "
+          "_stp_key_get_str ("+ qname() + ", " + stringify(i+1) + "); "
+          "if (! v) v = \"\"; "
+          "v; })";
       default:
 	throw semantic_error("illegal key type");
       }
@@ -510,10 +541,16 @@ c_unparser::emit_common_header ()
   o->newline() << "atomic_t session_state = ATOMIC_INIT (STAP_SESSION_STARTING);";
   o->newline();
   o->newline() << "struct context {";
-  o->newline(1) << "unsigned busy;";
+  o->newline(1) << "unsigned busy;"; // XXX: should be atomic_t ?
   o->newline() << "unsigned actioncount;";
-  o->newline() << "unsigned errorcount;";
   o->newline() << "unsigned nesting;";
+  o->newline() << "const char *last_error;";
+  // NB: last_error is used as a health flag within a probe.
+  // While it's 0, execution continues
+  // When it's "", current function or probe unwinds and returns early
+  // When it's "something", probe code unwinds, _stp_error's, sets error state
+  // See c_unparser::visit_statement()
+  o->newline() << "const char *last_stmt;";
   o->newline() << "struct pt_regs *regs;";
   o->newline() << "union {";
   o->indent(1);
@@ -582,11 +619,8 @@ c_unparser::emit_global (vardecl *v)
   else
     o->newline() << "static MAP global_" 
 		 << c_varname(v->name) << ";";
-  /* XXX
-  o->line() << "static DEFINE_RWLOCK("
-             << c_varname (v->name) << "_lock"
-             << ");";
-  */
+  o->newline() << "static rwlock_t "
+               << "global_" << c_varname (v->name) << "_lock;";
 }
 
 
@@ -617,6 +651,7 @@ c_unparser::emit_module_init ()
 	o->newline() << getmap (v).init();
       else
 	o->newline() << getvar (v).init();
+      o->newline() << "rwlock_init (& global_" << c_varname (v->name) << "_lock);";
     }
 
   for (unsigned i=0; i<session->probes.size(); i++)
@@ -624,7 +659,7 @@ c_unparser::emit_module_init ()
       o->newline() << "/* register " << i << " */";
       session->probes[i]->emit_registrations (o, i);
 
-      o->newline() << "if (rc) {";
+      o->newline() << "if (unlikely (rc)) {";
       // In case it's just a lower-layer (kprobes) error that set rc
       // but not session_state, do that here to prevent any other BEGIN
       // probe from attempting to run.
@@ -728,10 +763,12 @@ c_unparser::emit_function (functiondecl* v)
     << "& c->locals[c->nesting].function_" << c_varname (v->name)
     << ";";
   o->newline(-1) << "(void) l;"; // make sure "l" is marked used
+  o->newline() << "#define CONTEXT c";
   o->newline() << "#define THIS l";
   o->newline() << "if (0) goto out;"; // make sure out: is marked used
 
   // initialize locals
+  // XXX: optimization: use memset instead
   for (unsigned i=0; i<v->locals.size(); i++)
     {
       if (v->locals[i]->index_types.size() > 0) // array?
@@ -741,15 +778,20 @@ c_unparser::emit_function (functiondecl* v)
     }
 
   // initialize return value, if any
-  var retvalue = var(true, v->type, "__retvalue");
-  o->newline() << retvalue.init();
+  if (v->type != pe_unknown)
+    {
+      var retvalue = var(true, v->type, "__retvalue");
+      o->newline() << retvalue.init();
+    }
 
   v->body->visit (this);
+
   this->current_function = 0;
 
   o->newline(-1) << "out:";
   o->newline(1) << ";";
 
+  o->newline() << "#undef CONTEXT";
   o->newline() << "#undef THIS";
   o->newline(-1) << "}" << endl;
 }
@@ -921,7 +963,7 @@ c_unparser::c_typename (exp_type e)
 {
   switch (e)
     {
-    case pe_long: return string("long");
+    case pe_long: return string("int64_t");
     case pe_string: return string("string_t"); 
     case pe_stats: return string("stats_t");
     case pe_unknown: 
@@ -1046,12 +1088,13 @@ c_unparser_assignment::c_assignop(tmpvar & res,
       // - stats aggregation "<<<"
       // - modify-accumulate "+=" and many friends
       // - pre/post-crement "++"/"--"
+      // - "/" and "%" operators, but these need special handling in kernel
 
       // compute the modify portion of a modify-accumulate
       string macop;
       unsigned oplen = op.size();
       if (op == "=")
-	macop = "* 0 +"; // clever (?) trick to select rvalue (tmp) only
+	macop = "*error*"; // special shortcuts below
       else if (oplen > 1 && op[oplen-1] == '=') // for +=, %=, <<=, etc...
 	macop = op.substr(0, oplen-1);
       else if (op == "<<<")
@@ -1063,16 +1106,35 @@ c_unparser_assignment::c_assignop(tmpvar & res,
       else
 	// internal error
 	throw semantic_error ("unknown macop for assignment", tok);
-	
+
       if (pre)
 	{
+          if (macop == "/" || macop == "%" || op == "=")
+            throw semantic_error ("invalid pre-mode operator", tok);
+
 	  o->newline() << res << " = " << lval << ";";
-	  o->newline() << lval << " = " << res << " " << macop << " " << rval << ";";
+          o->newline() << lval << " = " << res << " " << macop << " " << rval << ";";
 	}
       else
 	{
-	  o->newline() << res << " = " << lval << " " << macop << " " << rval << ";";
-	  o->newline() << lval << " = " << res << ";";
+          if (op == "=") // shortcut simple assignment
+            {
+              o->newline() << lval << " = " << rval << ";";
+              res = rval;
+            }
+          else
+            {
+              if (macop == "/")
+                o->newline() << res << " = _stp_div64 (&c->last_error, "
+                             << lval << ", " << rval << ");";
+              else if (macop == "%")
+                o->newline() << res << " = _stp_mod64 (&c->last_error, "
+                             << lval << ", " << rval << ");";
+              else
+                o->newline() << res << " = " << lval << " " << macop << " " << rval << ";";
+
+              o->newline() << lval << " = " << res << ";";
+            }
 	}
     }
     else
@@ -1185,7 +1247,9 @@ c_unparser::getvar(vardecl *v, token const *tok)
 
 mapvar 
 c_unparser::getmap(vardecl *v, token const *tok) 
-{ 
+{   
+  if (v->arity < 1)
+    throw new semantic_error("attempt to use scalar where map expected", tok);
   return mapvar (is_local (v, tok), v->type, v->name, v->index_types);
 }
 
@@ -1197,20 +1261,48 @@ c_unparser::getiter(foreach_loop *f)
 }
 
 
+
+// An artificial common "header" for each statement.  This is where
+// activity counts limits and error state early exits are enforced.
+void
+c_unparser::visit_statement (statement *s, unsigned actions)
+{
+  // For some constructs, it is important to avoid an error branch
+  // right to the bottom of the probe/function.  The foreach() locking
+  // construct is one example.  Instead, if we are nested within a
+  // loop, we branch merely to its "break" label.  The next statement
+  // will branch one level higher, and so on, until we can go straight
+  // "out".
+  string outlabel = "out";
+  unsigned loops = loop_break_labels.size();
+  if (loops > 0)
+    outlabel = loop_break_labels[loops-1];
+
+  o->newline() << "if (unlikely (c->last_error)) goto " << outlabel << ";";
+  assert (s->tok);
+  o->newline() << "c->last_stmt = \"" << *s->tok << "\";";
+  if (actions > 0)
+    {
+      o->newline() << "c->actioncount += " << actions << ";";
+      o->newline() << "if (unlikely (c->actioncount > MAXACTION)) {";
+      o->newline(1) << "c->last_error = \"MAXACTION exceeded\";";
+      o->newline() << "goto " << outlabel << ";";
+      o->newline(-1) << "}";
+    }
+}
+
+
 void
 c_unparser::visit_block (block *s)
 {
   o->newline() << "{";
   o->indent (1);
-  o->newline() << "c->actioncount += " << s->statements.size() << ";";
-  o->newline() << "if (c->actioncount > MAXACTION) goto out;";
+  visit_statement (s, 0);
 
   for (unsigned i=0; i<s->statements.size(); i++)
     {
       try
         {
-          // XXX: it's probably not necessary to check this so frequently
-	  o->newline() << "if (c->errorcount) goto out;";
           s->statements[i]->visit (this);
 	  o->newline();
         }
@@ -1226,13 +1318,17 @@ c_unparser::visit_block (block *s)
 void
 c_unparser::visit_embeddedcode (embeddedcode *s)
 {
-  o->newline() << s->code;
+  visit_statement (s, 1);
+  o->newline() << "{";
+  o->newline(1) << s->code;
+  o->newline(-1) << "}";
 }
 
 
 void
 c_unparser::visit_null_statement (null_statement *s)
 {
+  visit_statement (s, 0);
   o->newline() << "/* null */;";
 }
 
@@ -1240,6 +1336,7 @@ c_unparser::visit_null_statement (null_statement *s)
 void
 c_unparser::visit_expr_statement (expr_statement *s)
 {
+  visit_statement (s, 1);
   o->newline() << "(void) ";
   s->value->visit (this);
   o->line() << ";";
@@ -1249,6 +1346,7 @@ c_unparser::visit_expr_statement (expr_statement *s)
 void
 c_unparser::visit_if_statement (if_statement *s)
 {
+  visit_statement (s, 1);
   o->newline() << "if (";
   o->indent (1);
   s->condition->visit (this);
@@ -1261,41 +1359,59 @@ c_unparser::visit_if_statement (if_statement *s)
     {
       o->newline() << "else {";
       o->indent (1);
-      s->thenblock->visit (this);
+      s->elseblock->visit (this);
       o->newline(-1) << "}";
     }
 }
 
 
 void
-c_unparser::visit_for_loop (for_loop *s)
+c_tmpcounter::visit_for_loop (for_loop *s)
 {
   s->init->visit (this);
+  s->cond->visit (this);
+  s->block->visit (this);
+  s->incr->visit (this);
+}
+
+
+void
+c_unparser::visit_for_loop (for_loop *s)
+{
+  visit_statement (s, 1);
+
   string ctr = stringify (label_counter++);
+  string toplabel = "top_" + ctr;
   string contlabel = "continue_" + ctr;
   string breaklabel = "break_" + ctr;
 
-  o->newline() << contlabel << ":";
+  // initialization
+  s->init->visit (this);
 
-  o->newline() << "c->actioncount ++;";
-  o->newline() << "if (c->actioncount > MAXACTION) goto out;";
-
-  o->newline() << "if (! (";
+  // condition
+  o->newline(-1) << toplabel << ":";
+  o->newline(1) << "if (! (";
   if (s->cond->type != pe_long)
     throw semantic_error ("expected numeric type", s->cond->tok);
   s->cond->visit (this);
   o->line() << ")) goto " << breaklabel << ";";
 
+  // body
   loop_break_labels.push_back (breaklabel);
   loop_continue_labels.push_back (contlabel);
   s->block->visit (this);
   loop_break_labels.pop_back ();
   loop_continue_labels.pop_back ();
 
+  // iteration
+  o->newline(-1) << contlabel << ":";
+  o->indent(1);
   s->incr->visit (this);
-  o->newline() << "goto " << contlabel << ";";
+  o->newline() << "goto " << toplabel << ";";
 
-  o->newline() << breaklabel << ": ; /* dummy statement */";
+  // exit
+  o->newline(-1) << breaklabel << ":";
+  o->newline(1) << "; /* dummy statement */";
 }
 
 
@@ -1310,36 +1426,60 @@ c_tmpcounter::visit_foreach_loop (foreach_loop *s)
 void
 c_unparser::visit_foreach_loop (foreach_loop *s)
 {
+  visit_statement (s, 1);
+
   mapvar mv = getmap (s->base_referent, s->tok);
   itervar iv = getiter (s);
   vector<var> keys;
-    
-  varlock guard (*this, mv);
-    
-  o->newline() << "for ("
-	       << iv << " = " << iv.start (mv) << "; ";
-  o->newline() << "     " << iv << "; ";
-  o->newline() << "     " << iv << " = " << iv.next (mv) << ")";
+
+  string ctr = stringify (label_counter++);
+  string toplabel = "top_" + ctr;
+  string contlabel = "continue_" + ctr;
+  string breaklabel = "break_" + ctr;
+
+  // NB: structure parallels for_loop
+
+  // initialization
+  varlock_r guard (*this, mv);
+  o->newline() << iv << " = " << iv.start (mv) << ";";
+
+  // condition
+  o->newline(-1) << toplabel << ":";
+  o->newline(1) << "if (! (" << iv << ")) goto " << breaklabel << ";";
+
+  // body
+  loop_break_labels.push_back (breaklabel);
+  loop_continue_labels.push_back (contlabel);
   o->newline() << "{";
   o->indent (1);
-    
   for (unsigned i = 0; i < s->indexes.size(); ++i)
     {
       // copy the iter values into the specified locals
       var v = getvar (s->indexes[i]->referent);
       c_assign (v, iv.get_key (v.type(), i), s->tok);
     }
-    
   s->block->visit (this);
+  o->newline(-1) << "}";
+  loop_break_labels.pop_back ();
+  loop_continue_labels.pop_back ();
+
+  // iteration
+  o->newline(-1) << contlabel << ":";
+  o->newline(1) << iv << " = " << iv.next (mv) << ";";
+  o->newline() << "goto " << toplabel << ";";
     
-  o->indent (-1);
-  o->newline() << "}";
+  // exit
+  o->newline(-1) << breaklabel << ":";
+  o->indent(1);
+  // varlock dtor will show up here
 }
 
 
 void
 c_unparser::visit_return_statement (return_statement* s)
 {
+  visit_statement (s, 1);
+
   if (current_function == 0)
     throw semantic_error ("cannot 'return' from probe", s->tok);
 
@@ -1348,30 +1488,73 @@ c_unparser::visit_return_statement (return_statement* s)
                          "vs", s->tok);
 
   c_assign ("l->__retvalue", s->value, "return value");
-  o->newline() << "goto out;";
+  o->newline() << "c->last_error = \"\";";
+  // NB: last_error needs to get reset to NULL in the caller
+  // probe/function
 }
 
 
 void
 c_unparser::visit_next_statement (next_statement* s)
 {
+  visit_statement (s, 1);
+
   if (current_probe == 0)
     throw semantic_error ("cannot 'next' from function", s->tok);
 
-  o->newline() << "goto out;";
+  o->newline() << "c->last_error = \"\";";
+}
+
+
+struct delete_statement_operand_visitor:
+  public throwing_visitor
+{
+  c_unparser *parent;
+  delete_statement_operand_visitor (c_unparser *p):
+    throwing_visitor ("invalid operand of delete expression"),
+    parent (p)
+  {}
+  void visit_symbol (symbol* e);
+  void visit_arrayindex (arrayindex* e);
+};
+
+void 
+delete_statement_operand_visitor::visit_symbol (symbol* e)
+{
+  mapvar mvar = parent->getmap(e->referent, e->tok);  
+  varlock_w guard (*parent, mvar);
+  parent->o->newline() << mvar.fini ();
+  parent->o->newline() << mvar.init ();  
+}
+
+void 
+delete_statement_operand_visitor::visit_arrayindex (arrayindex* e)
+{
+  vector<tmpvar> idx;
+  parent->load_map_indices (e, idx);
+
+  {
+    mapvar mvar = parent->getmap (e->referent, e->tok);
+    varlock_w guard (*parent, mvar);
+    parent->o->newline() << mvar.seek (idx) << ";";
+    parent->o->newline() << mvar.del () << ";";
+  }
 }
 
 
 void
 c_unparser::visit_delete_statement (delete_statement* s)
 {
-  throw semantic_error ("delete statement not yet implemented", s->tok);
+  visit_statement (s, 1);
+  delete_statement_operand_visitor dv (this);
+  s->value->visit (&dv);
 }
 
 
 void
 c_unparser::visit_break_statement (break_statement* s)
 {
+  visit_statement (s, 1);
   if (loop_break_labels.size() == 0)
     throw semantic_error ("cannot 'break' outside loop", s->tok);
 
@@ -1383,6 +1566,7 @@ c_unparser::visit_break_statement (break_statement* s)
 void
 c_unparser::visit_continue_statement (continue_statement* s)
 {
+  visit_statement (s, 1);
   if (loop_continue_labels.size() == 0)
     throw semantic_error ("cannot 'continue' outside loop", s->tok);
 
@@ -1402,7 +1586,9 @@ c_unparser::visit_literal_string (literal_string* e)
 void
 c_unparser::visit_literal_number (literal_number* e)
 {
-  o->line() << e->value;
+  // This looks ugly, but tries to be warning-free on 32- and 64-bit
+  // hosts.
+  o->line() << "((uint64_t)" << e->value << "LL)";
 }
 
 
@@ -1463,9 +1649,9 @@ c_unparser::visit_binary_expression (binary_expression* e)
       e->right->visit (this);
       o->line() << ";";
 
-      right.check_dbz(*this);
+      o->newline() << ((e->op == "/") ? "_stp_div64" : "_stp_mod64")
+                   << " (&c->last_error, " << left << ", " << right << ");";
 
-      o->newline() << left << " " << e->op << " " << right << ";";
       o->newline(-1) << "})";
     }
   else
@@ -1517,10 +1703,43 @@ c_unparser::visit_logical_and_expr (logical_and_expr* e)
 }
 
 
+void 
+c_tmpcounter::visit_array_in (array_in* e)
+{
+  vardecl* r = e->operand->referent;
+
+  // One temporary per index dimension.
+  for (unsigned i=0; i<r->index_types.size(); i++)
+    {
+      tmpvar ix = parent->gensym (r->index_types[i]);
+      ix.declare (*parent);
+      e->operand->indexes[i]->visit(this);
+    }
+ 
+ // A boolean result.
+  tmpvar res = parent->gensym (e->type);
+  res.declare (*parent);
+}
+
+
 void
 c_unparser::visit_array_in (array_in* e)
 {
-  throw semantic_error ("array-in expression not yet implemented", e->tok);
+  stmt_expr block(*this);  
+
+  vector<tmpvar> idx;
+  load_map_indices (e->operand, idx);
+
+  tmpvar res = gensym (pe_long);
+
+  { // block used to control varlock_r lifespan
+    mapvar mvar = getmap (e->operand->referent, e->tok);
+    varlock_r guard (*this, mvar);
+    o->newline() << mvar.seek (idx) << ";";
+    c_assign (res, mvar.exists(), e->tok);
+  }
+
+  o->newline() << res << ";";
 }
 
 
@@ -1746,9 +1965,6 @@ c_unparser_assignment::prepare_rvalue (string const & op,
         throw semantic_error ("need rvalue for assignment", tok);
     }
   // OPT: literal rvalues could be used without a tmp* copy
-
-  if (op == "/=" || op == "%=") 
-    rval.check_dbz (*parent);
 }
 
 void
@@ -1766,7 +1982,7 @@ c_unparser_assignment::visit_symbol (symbol *e)
 
   {
     var lvar = parent->getvar (e->referent, e->tok);
-    varlock guard (*parent, lvar);
+    varlock_w guard (*parent, lvar);
     c_assignop (res, lvar, rval, e->tok);     
   }
 
@@ -1774,6 +1990,13 @@ c_unparser_assignment::visit_symbol (symbol *e)
 }
 
 
+void 
+c_unparser::visit_target_symbol (target_symbol* e)
+{
+  throw semantic_error("cannot translate general target-symbol expression");
+}
+
+
 void
 c_tmpcounter::visit_arrayindex (arrayindex *e)
 {
@@ -1843,11 +2066,9 @@ c_unparser::visit_arrayindex (arrayindex* e)
   // reentrancy issues that pop up with nested expressions:
   // e.g. a[a[c]=5] could deadlock
   
-  o->newline() << "if (c->errorcount) goto out;";
-  
-  {
+  { // block used to control varlock_r lifespan
     mapvar mvar = getmap (e->referent, e->tok);
-    varlock guard (*this, mvar);
+    varlock_r guard (*this, mvar);
     o->newline() << mvar.seek (idx) << ";";
     c_assign (res, mvar.get(), e->tok);
   }
@@ -1908,8 +2129,8 @@ c_unparser_assignment::visit_arrayindex (arrayindex *e)
   // thusly:
   // ({ tmp0=(idx0); ... tmpN=(idxN); rvar=(rhs); lvar; res;
   //    rvar = ...;
-  //    lock (array);n
-  //    lvar = get (array,idx0...N);
+  //    lock (array);
+  //    lvar = get (array,idx0...N); // if necessary
   //    assignop (res, lvar, rvar);
   //    set (array, idx0...N, lvar);
   //    unlock (array);
@@ -1919,15 +2140,14 @@ c_unparser_assignment::visit_arrayindex (arrayindex *e)
   // reentrancy issues that pop up with nested expressions:
   // e.g. ++a[a[c]=5] could deadlock
   
-  o->newline() << "if (c->errorcount) goto out;";
-  
   prepare_rvalue (op, rvar, e->tok);
   
-  {
+  { // block used to control varlock_w lifespan
     mapvar mvar = parent->getmap (e->referent, e->tok);
-    varlock guard (*parent, mvar);
+    varlock_w guard (*parent, mvar);
     o->newline() << mvar.seek (idx) << ";";
-    parent->c_assign (lvar, mvar.get(), e->tok);
+    if (op != "=") // don't bother fetch slot if we will just overwrite it
+      parent->c_assign (lvar, mvar.get(), e->tok);
     c_assignop (res, lvar, rvar, e->tok); 
     o->newline() << mvar.set (lvar) << ";";
   }
@@ -1982,10 +2202,10 @@ c_unparser::visit_functioncall (functioncall* e)
     }
 
   o->newline();
-  o->newline() << "if (c->nesting+2 >= MAXNESTING) goto out;";
-  o->newline() << "c->actioncount ++;";
-  o->newline() << "if (c->actioncount > MAXACTION) goto out;";
-  o->newline();
+  o->newline() << "if (unlikely (c->nesting+2 >= MAXNESTING)) {";
+  o->newline(1) << "c->last_error = \"MAXNESTING exceeded\";";
+  o->newline(-1) << "} else {";
+  o->indent(1);
 
   // copy in actual arguments
   for (unsigned i=0; i<e->args.size(); i++)
@@ -2007,7 +2227,14 @@ c_unparser::visit_functioncall (functioncall* e)
   o->newline() << "c->nesting ++;";
   o->newline() << "function_" << c_varname (r->name) << " (c);";
   o->newline() << "c->nesting --;";
-  
+
+  // reset last_error to NULL if it was set to "" by return()
+  o->newline() << "if (c->last_error && ! c->last_error[0])";
+  o->newline(1) << "c->last_error = 0;";
+  o->indent(-1);
+
+  o->newline(-1) << "}";
+
   // return result from retvalue slot
   
   if (r->type == pe_unknown)
@@ -2038,6 +2265,7 @@ translate_pass (systemtap_session& s)
       s.op->newline() << "#else";
       s.op->newline() << "#include \"runtime.h\"";
       s.op->newline() << "#include <linux/string.h>";
+      s.op->newline() << "#include <linux/timer.h>";
       // XXX
       s.op->newline() << "#define KALLSYMS_LOOKUP_NAME \"\"";
       s.op->newline() << "#define KALLSYMS_LOOKUP 0";
